from __future__ import annotations

from collections.abc import Callable
from ctypes import CDLL, c_int, c_void_p
from ctypes.util import find_library
from datetime import datetime
from errno import ENOENT, EROFS
from functools import wraps
import io
import logging
import os
import os.path as op
from pathlib import Path
import stat
import sys
from threading import Lock
from typing import IO, Any, Optional, TypeVar

from datalad import cfg
from datalad.distribution.dataset import Dataset
from fuse import FuseOSError, Operations
import methodtools

from .consts import CACHE_SIZE
from .fsspec import FsspecAdapter

# Make it relatively small since we are aiming for metadata records ATM
# Seems of no real good positive net ATM
# BLOCK_SIZE = 2**20  # 1M. block size to fetch at a time.
from .utils import AnnexDir, AnnexKey, is_annex_dir_or_key

if sys.version_info[:2] >= (3, 10):
    from typing import Concatenate, ParamSpec
else:
    from typing_extensions import Concatenate, ParamSpec

lgr = logging.getLogger("datalad.fuse")

libcname = find_library("c")
assert libcname is not None
libc = CDLL(libcname)

fcntl = libc.fcntl
fcntl.argtypes = [c_int, c_int, c_void_p]
fcntl.restype = c_int

T = TypeVar("T")
P = ParamSpec("P")


def write_op(
    f: Callable[Concatenate[DataLadFUSE, str, P], T]
) -> Callable[Concatenate[DataLadFUSE, str, P], T]:
    """Decorator for operations which need to write"""

    @wraps(f)
    def wrapped(self: DataLadFUSE, path: str, *args: P.args, **kwargs: P.kwargs) -> T:
        if self.mode_transparent and self.is_under_git(path):
            return f(self, path, *args, **kwargs)
        else:
            raise FuseOSError(EROFS)

    return wrapped


class DataLadFUSE(Operations):  # LoggingMixIn,
    # ??? TODO: since we would mix normal os.open
    # and not, we will mint our "fds" over this offset

    _counter_offset = 1000

    def __init__(self, root: str, mode_transparent: bool = False) -> None:
        self.root = op.realpath(root)
        self.mode_transparent = mode_transparent
        self.rwlock = Lock()
        self._adapter = FsspecAdapter(root, mode_transparent=mode_transparent)
        self._fhdict: dict[int, Optional[IO[bytes]]] = {}
        # fh to fsspec_file, already opened (we are RO for now, so can just open
        # and there is no seek so we should be ok even if the same file open
        # multiple times?
        self._counter = DataLadFUSE._counter_offset

    def __call__(self, op: str, path: str, *args: Any) -> Any:
        lgr.debug("op=%s for path=%s with args %s", op, path, args)
        # if (".git", "annex", "objects") == Path(path).parts[-7:-4]:
        #     import pdb; pdb.set_trace()
        if not self.mode_transparent and ".git" in Path(path).parts:
            lgr.debug("Raising ENOENT for .git")
            raise FuseOSError(ENOENT)
        return super(DataLadFUSE, self).__call__(op, self.root + path, *args)

    def destroy(self, _path: Optional[str] = None) -> int:
        lgr.warning("Destroying fsspecs and collection of %d fhs", len(self._fhdict))
        for f in self._fhdict.values():
            if f is not None:
                try:
                    f.close()
                except Exception as e:
                    lgr.error("%s", e)
        self._fhdict = {}
        cache_clear = cfg.get("datalad.fusefs.cache-clear")
        if cache_clear == "visited":
            for dsap in self._adapter.datasets.values():
                dsap.clear()
        elif cache_clear == "recursive":
            Dataset(self.root).fsspec_cache_clear(recursive=True)
        return 0

    @staticmethod
    # XXX not yet sure what we need to filter...
    def _filter_stat(st: os.stat_result) -> dict[str, Any]:
        return dict(
            (key, getattr(st, key))
            for key in (
                "st_atime",
                "st_ctime",
                "st_gid",
                "st_mode",
                "st_mtime",
                "st_nlink",
                "st_size",
                "st_uid",
            )
        )

    @methodtools.lru_cache(maxsize=CACHE_SIZE)
    def getattr(self, path: str, fh: Optional[int] = None) -> dict[str, Any]:
        # TODO: support of unlocked files... but at what cost?
        lgr.debug("getattr(path=%r, fh=%r)", path, fh)
        r: Optional[dict[str, Any]] = None
        if fh and fh < self._counter_offset:
            lgr.debug("Calling os.fstat()")
            r = self._filter_stat(os.fstat(fh))
        elif op.exists(path):
            lgr.debug("File exists; calling os.stat()")
            r = self._filter_stat(os.stat(path))
        elif self.mode_transparent:
            if op.lexists(path):
                lgr.debug("Broken symlink; calling os.lstat()")
                r = self._filter_stat(os.lstat(path))
            else:
                iadok = is_annex_dir_or_key(path)
                if iadok is not None:
                    if isinstance(iadok, AnnexKey):
                        if iadok.size is not None:
                            lgr.debug("Got size from key")
                            r = mkstat(
                                is_file=True,
                                size=iadok.size,
                                timestamp=self._adapter.get_commit_datetime(path),
                            )
                        else:
                            # needs to be open but it is a key. We will let
                            # fsspec handle it
                            pass
                    elif isinstance(iadok, AnnexDir):
                        # just return that one of the top directory
                        # TODO: cache this since would be a frequent operation
                        r = self._filter_stat(os.stat(iadok.topdir))
                    else:
                        raise AssertionError(f"Unexpected iadok: {iadok!r}")
                elif self.is_under_git(path):
                    lgr.debug("Path under .git does not exist; raising ENOENT")
                    raise FuseOSError(ENOENT)
        if r is None:
            fsspec_file = None
            if fh and fh >= self._counter_offset:
                lgr.debug("File already open")
                fsspec_file = self._fhdict[fh]
                to_close = False
            else:
                _, key = self._adapter.get_file_state(path)
                assert key is not None
                if key.size is not None:
                    lgr.debug("Got size from key")
                    r = mkstat(
                        is_file=True,
                        size=key.size,
                        timestamp=self._adapter.get_commit_datetime(path),
                    )
                else:
                    lgr.debug("File not already open")
                    with self.rwlock:
                        fsspec_file = self._adapter.open(path)
                    to_close = True
            if fsspec_file is not None:
                if isinstance(fsspec_file, io.BufferedIOBase):
                    # full file was already fetched locally
                    lgr.debug("File object is io.BufferedIOBase")
                    r = self._filter_stat(os.stat(fsspec_file.name))
                else:
                    lgr.debug("File object is fsspec object")
                    r = file_getattr(
                        fsspec_file, timestamp=self._adapter.get_commit_datetime(path)
                    )
                if to_close:
                    with self.rwlock:
                        fsspec_file.close()
        lgr.debug("Returning %r for %s", r, path)
        assert r is not None
        return r

    def open(self, path: str, flags: int) -> int:
        lgr.debug("open(path=%r, flags=%#x)", path, flags)
        # fn = "".join([self.root, path.lstrip("/")])
        if op.exists(path) or (
            self.mode_transparent
            and self.is_under_git(path)
            and is_annex_dir_or_key(path) is None
        ):
            if op.exists(path):
                lgr.debug("Path exists; opening directly")
            else:
                lgr.debug("Path is under .git/; opening directly")
            fh = os.open(path, flags)
            if fh >= self._counter_offset:
                raise RuntimeError(
                    "We got file handle %d, our hopes that we never get such"
                    " high one were wrong" % fh
                )
            return fh
        else:
            lgr.debug("Opening path via fsspec")
            if flags % 2 == 0:
                # read
                mode = "rb"  # noqa: F841
            else:
                # write/create
                raise FuseOSError(EROFS)
            with self.rwlock:
                fsspec_file = self._adapter.open(path)
            lgr.debug("Counter = %d", self._counter)
            # TODO: threadlock ?
            self._fhdict[self._counter] = fsspec_file  # self.fs.open(fn, mode)
            self._counter += 1
            return self._counter - 1

    def read(self, _path: str, size: int, offset: int, fh: int) -> bytes:
        lgr.debug("read(path=%r, size=%r, offset=%r, fh=%r)", _path, size, offset, fh)
        if fh < self._counter_offset:
            lgr.debug("Reading directly")
            with self.rwlock:
                os.lseek(fh, offset, 0)
                return os.read(fh, size)
        else:
            lgr.debug("Reading from open filehandle")
            # must be open already and we must have mapped it to fsspec file
            # TODO: check for path to correspond?
            f = self._fhdict[fh]
            assert f is not None
            with self.rwlock:
                f.seek(offset)
                return f.read(size)

    def opendir(self, path: str) -> int:
        lgr.debug("opendir(path=%r)", path)
        if not op.exists(path):
            lgr.debug("Directory does not exist; raising ENOENT")
            raise FuseOSError(ENOENT)
        lgr.debug("Counter = %d", self._counter)
        # TODO: threadlock ?
        self._fhdict[self._counter] = None
        self._counter += 1
        return self._counter - 1

    def readdir(self, path: str, _fh: int) -> list[str]:
        lgr.debug("readdir(path=%r, fh=%r)", path, _fh)
        paths = [".", ".."] + os.listdir(path)
        if not self.mode_transparent:
            try:
                paths.remove(".git")
            except ValueError:
                pass
            else:
                lgr.debug("Removed .git from dirlist")
        return paths

    def release(self, path: str, fh: int) -> int:
        lgr.debug("release(path=%r, fh=%r)", path, fh)
        if fh < self._counter_offset:
            lgr.debug("Closing directly")
            os.close(fh)
        elif fh in self._fhdict:
            lgr.debug("Popping from filehandle collection")
            f = self._fhdict.pop(fh)
            # but we do not close an fsspec instance, so it could be reused
            # on subsequent accesses
            # TODO: this .close is not sufficient -- _fhdict is breeding open
            #  files, so we need to provide some proper use of lru_cache
            #  to have not recently used closed
            if f is not None and not f.closed:
                with self.rwlock:
                    f.close()
        return 0

    def readlink(self, path: str) -> str:
        lgr.debug("readlink(path=%r)", path)
        return os.readlink(path)

    # ??? seek seems to be not implemented by fusepy/ Operations

    #
    # Benign writeable operations which we can allow
    #
    mkdir = os.mkdir
    mknod = os.mknod
    rmdir = os.rmdir
    chmod = os.chmod
    chown = os.chown

    def flush(self, _path: str, fh: int) -> None:
        lgr.debug("flush(path=%r, fh=%r)", _path, fh)
        if fh < self._counter_offset:
            lgr.debug("Flushing directly")
            os.fsync(fh)

    def fsync(self, _path: str, datasync: int, fh: int) -> None:
        lgr.debug("fsync(path=%r, datasync=%r, fh=%r)", _path, datasync, fh)
        if fh < self._counter_offset:
            lgr.debug("Fsyncing directly")
            if datasync != 0:
                os.fdatasync(fh)  # type: ignore[attr-defined]
            else:
                os.fsync(fh)

    #
    # Extra operations we do not have implemented
    #
    getxattr = None
    listxattr = None

    ioctl = None

    # def access(self, path, mode):
    #     if not os.access(path, mode):
    #         raise FuseOSError(EACCES)
    #

    @write_op
    def create(self, path: str, mode: int) -> int:
        return os.open(path, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, mode)

    @write_op
    def link(self, target: str, source: str) -> None:
        if "/.git/" in source:
            os.link(self.root + source, target)
        else:
            raise FuseOSError(EROFS)

    @write_op
    def rename(self, old: str, new: str) -> None:
        if "/.git/" in new:
            os.rename(old, self.root + new)
        else:
            raise FuseOSError(EROFS)

    # def statfs(self, path: str) -> dict[str, Any]:
    #     lgr.mydebug(f"statfs {path}")
    #     raise NotImplementedError()
    #     stv = os.statvfs(path)
    #     return dict((key, getattr(stv, key)) for key in (
    #         'f_bavail', 'f_bfree', 'f_blocks', 'f_bsize', 'f_favail',
    #         'f_ffree', 'f_files', 'f_flag', 'f_frsize', 'f_namemax'))

    @write_op
    def symlink(self, target: str, source: str) -> None:
        os.symlink(source, target)

    @write_op
    def truncate(self, path: str, length: int, _fh: Optional[int] = None) -> None:
        with open(path, "r+") as f:
            f.truncate(length)

    @write_op
    def unlink(self, path: str) -> None:
        with self.rwlock:
            os.unlink(path)

    def utimens(self, path: str, times: Optional[tuple[int, int]] = None) -> None:
        if times is not None:
            os.utime(path, ns=times)
        else:
            os.utime(path)

    @write_op
    def write(self, _path: str, data: bytes, offset: int, fh: int) -> int:
        with self.rwlock:
            os.lseek(fh, offset, 0)
            return os.write(fh, data)

    @write_op
    def lock(self, _path: str, fh: int, cmd: int, lock: int) -> int:
        r = fcntl(fh, cmd, lock)
        assert isinstance(r, int)
        return r

    def is_under_git(self, path: str) -> bool:
        return ".git" in Path(path).relative_to(self.root).parts


def file_getattr(f: Any, timestamp: datetime) -> dict[str, Any]:
    # code borrowed from fsspec.fuse:FUSEr.getattr
    # TODO: improve upon! there might be mtime of url
    try:
        info = f.info()
    except FileNotFoundError:
        raise FuseOSError(ENOENT)
    return mkstat(info["type"] == "file", info["size"], timestamp)


def mkstat(is_file: bool, size: int, timestamp: datetime) -> dict[str, Any]:
    # TODO Also I get UID.GID funny -- yarik, not yoh
    # get of the original symlink, so float it up!
    data: dict[str, Any] = {"st_uid": os.getuid(), "st_gid": os.getgid()}
    if not is_file:
        data["st_mode"] = stat.S_IFDIR | 0o755
        data["st_size"] = 0
        data["st_blksize"] = 0
    else:
        data["st_mode"] = stat.S_IFREG | 0o644
        data["st_size"] = size
        data["st_blksize"] = 5 * 2**20
        data["st_nlink"] = 1
    data["st_atime"] = timestamp.timestamp()
    data["st_ctime"] = timestamp.timestamp()
    data["st_mtime"] = timestamp.timestamp()
    return data



from pathlib import Path
import subprocess

from datalad.api import Dataset
from datalad.tests.utils_pytest import assert_in_results
from linesep import split_terminated
import pytest


def first_n_lines(blob, n):
    # In order to match the behavior of Python's binary IO files, only \n
    # should be treated as a line separator, so bytes.splitlines() is not an
    # option.
    return b"".join(split_terminated(blob, b"\n", True)[:n])


def test_get_default_lines_text(url_dataset):
    ds, data_files = url_dataset
    assert_in_results(
        ds.fsspec_head("text.txt"),
        action="fsspec-head",
        type="dataset",
        status="ok",
        data=first_n_lines(data_files["text.txt"], 10),
    )


@pytest.mark.parametrize("lines", [0, 4, 20])
def test_get_lines_text(lines, url_dataset):
    ds, data_files = url_dataset
    assert_in_results(
        ds.fsspec_head("text.txt", lines=lines),
        action="fsspec-head",
        type="dataset",
        status="ok",
        data=first_n_lines(data_files["text.txt"], lines),
    )


def test_get_bytes_text(url_dataset):
    ds, data_files = url_dataset
    assert_in_results(
        ds.fsspec_head("text.txt", bytes=100),
        action="fsspec-head",
        type="dataset",
        status="ok",
        data=data_files["text.txt"][:100],
    )


def test_get_lines_binary(url_dataset):
    ds, data_files = url_dataset
    assert_in_results(
        ds.fsspec_head("binary.png", lines=3),
        action="fsspec-head",
        type="dataset",
        status="ok",
        data=first_n_lines(data_files["binary.png"], 3),
    )


def test_get_bytes_binary(url_dataset):
    ds, data_files = url_dataset
    assert_in_results(
        ds.fsspec_head("binary.png", bytes=100),
        action="fsspec-head",
        type="dataset",
        status="ok",
        data=data_files["binary.png"][:100],
    )


def test_subdataset_get_default_lines_text(superdataset):
    ds, data_files = superdataset
    assert_in_results(
        ds.fsspec_head("sub/text.txt"),
        action="fsspec-head",
        type="dataset",
        status="ok",
        data=first_n_lines(data_files["sub/text.txt"], 10),
    )


@pytest.mark.parametrize("lines", [0, 4, 20])
def test_subdataset_get_lines_text(lines, superdataset):
    ds, data_files = superdataset
    assert_in_results(
        ds.fsspec_head("sub/text.txt", lines=lines),
        action="fsspec-head",
        type="dataset",
        status="ok",
        data=first_n_lines(data_files["sub/text.txt"], lines),
    )


def test_subdataset_get_bytes_text(superdataset):
    ds, data_files = superdataset
    assert_in_results(
        ds.fsspec_head("sub/text.txt", bytes=100),
        action="fsspec-head",
        type="dataset",
        status="ok",
        data=data_files["sub/text.txt"][:100],
    )


def test_subdataset_get_lines_binary(superdataset):
    ds, data_files = superdataset
    assert_in_results(
        ds.fsspec_head("sub/binary.png", lines=3),
        action="fsspec-head",
        type="dataset",
        status="ok",
        data=first_n_lines(data_files["sub/binary.png"], 3),
    )


def test_subdataset_get_bytes_binary(superdataset):
    ds, data_files = superdataset
    assert_in_results(
        ds.fsspec_head("sub/binary.png", bytes=100),
        action="fsspec-head",
        type="dataset",
        status="ok",
        data=data_files["sub/binary.png"][:100],
    )


def test_git_repo(tmp_path):
    subprocess.run(["git", "init"], cwd=tmp_path, check=True)
    TEXT = (Path(__file__).with_name("data") / "text.txt").read_bytes()
    (tmp_path / "text.txt").write_bytes(TEXT)
    subprocess.run(["git", "add", "text.txt"], cwd=tmp_path, check=True)
    subprocess.run(["git", "commit", "-m", "Add a file"], cwd=tmp_path, check=True)
    ds = Dataset(tmp_path)
    assert_in_results(
        ds.fsspec_head("text.txt"),
        action="fsspec-head",
        type="dataset",
        status="ok",
        data=first_n_lines(TEXT, 10),
    )



from contextlib import contextmanager
from dataclasses import dataclass
from http.server import HTTPServer, SimpleHTTPRequestHandler
import logging
import multiprocessing
import os
import os.path
from pathlib import Path
import re
import time
from typing import List

from datalad.api import Dataset, clone
import pytest
import requests

DATA_DIR = Path(__file__).with_name("data")

lgr = logging.getLogger("datalad.fuse.tests")

# We need large files that exceed the blocksize of 1 MiB
BIG_URLS = [
    # (dataset path, URL, SHA256 digest)
    (
        "APL.pdf",
        "http://www.softwarepreservation.org/projects/apl/Books/APROGRAMMING%20LANGUAGE",
        "c65ccc2a97cdb6042641847112dc6e4d4d6e75fdedcd476fdd61f855711bbaf4",
    ),
    (
        "gameboy.pdf",
        "https://archive.org/download/GameBoyProgManVer1.1/GameBoyProgManVer1.1.pdf",
        "5263e6c1f5fa51fc6813d2ed71c738c887ec78554eef633ad72c6285f7ff9197",
    ),
    (
        "libpython3.10-stdlib_3.10.4-3_i386.deb",
        "http://nyc3.clouds.archive.ubuntu.com/ubuntu/pool/main/p/python3.10/libpython3.10-stdlib_3.10.4-3_i386.deb",
        "e79c1416ec792b61ad9770f855bf6889e57be5f6511ea814d81ef5f9b1a3eec9",
    ),
]


@pytest.fixture(autouse=True)
def capture_all_logs(caplog):
    caplog.set_level(logging.DEBUG, logger="datalad.fuse")


def pytest_addoption(parser) -> None:
    parser.addoption(
        "--libfuse",
        action="store_true",
        default=False,
        help="Enable fuse tests",
    )


def pytest_collection_modifyitems(config, items):
    if not config.getoption("--libfuse"):
        skip_no_libfuse = pytest.mark.skip(reason="Only run when --libfuse is given")
        for item in items:
            if "libfuse" in item.keywords:
                item.add_marker(skip_no_libfuse)


@pytest.fixture()
def tmp_home(monkeypatch, tmp_path_factory):
    home = tmp_path_factory.mktemp("tmp_home")
    monkeypatch.setenv("HOME", str(home))
    monkeypatch.setenv("USERPROFILE", str(home))
    monkeypatch.delenv("XDG_CONFIG_HOME", raising=False)
    monkeypatch.setenv("LOCALAPPDATA", str(home))
    (home / ".gitconfig").write_text(
        '[annex "security"]\n'
        "allowed-url-schemes = http https file\n"
        "allowed-http-addresses = all\n"
        "\n"
        "[user]\n"
        "name = Test User\n"
        "email = test@test.nil\n"
    )
    return home


def serve_path_via_http(hostname, path, queue):
    os.chdir(path)
    httpd = HTTPServer((hostname, 0), SimpleHTTPRequestHandler)
    queue.put(httpd.server_port)
    httpd.serve_forever()


@contextmanager
def local_server(directory):
    hostname = "127.0.0.1"
    queue = multiprocessing.Queue()
    p = multiprocessing.Process(
        target=serve_path_via_http, args=(hostname, directory, queue)
    )
    p.start()
    try:
        port = queue.get(timeout=300)
        url = f"http://{hostname}:{port}"
        lgr.debug("HTTP: serving %s at %s", directory, url)
        with pytest.MonkeyPatch().context() as m:
            m.delenv("http_proxy", raising=False)
            for _ in range(10):
                try:
                    requests.get(url, timeout=1)
                except requests.RequestException:
                    time.sleep(0.1)
                else:
                    break
            else:
                raise RuntimeError("Server did not come up in time")
            yield url
    finally:
        lgr.debug("HTTP: stopping server")
        p.terminate()


@dataclass
class DataFile:
    url: str
    path: str
    content: bytes


@pytest.fixture(scope="session")
def served_files():
    with local_server(DATA_DIR) as url:
        files = []
        for p in DATA_DIR.iterdir():
            if p.is_file():
                files.append(
                    DataFile(
                        url=f"{url}/{p.name}",
                        path=p.name,
                        content=p.read_bytes(),
                    )
                )
        yield files


def initdataset(ds, data_files: List[DataFile], is_remote: bool) -> None:
    for dfile in data_files:
        if is_remote:
            ds.repo.add_url_to_file(dfile.path, dfile.url, options=["--relaxed"])
        else:
            ds.download_url(urls=dfile.url, path=dfile.path)
        # Add an invalid URL in order to test bad-URL-fallback.
        # Add it after `download_url()` is called in order to not cause an
        #  error on filesystems without symlinks (in which case doing
        #  `add_url_to_file()` would cause the file to be created as a
        #  non-symlink, which `download_url()` would then see as the file
        #  already being present, leading to an error).
        # It appears that git annex returns files' URLs in lexicographic order,
        #  so in order for the bad URL to be tried first, we insert a '0' at
        #  the start of the URL path.
        ds.repo.add_url_to_file(
            dfile.path, re.sub(r"(:\d+/)", r"\g<1>0", dfile.url), options=["--relaxed"]
        )


# @pytest.mark.usefixtures("tmp_home")  # Doesn't work on fixture functions
@pytest.fixture(params=["remote", "local", "cloned"])
def url_dataset(served_files, request, tmp_home, tmp_path_factory):  # noqa: U100
    workpath = tmp_path_factory.mktemp("url_dataset")
    ds = Dataset(workpath / "ds").create()
    initdataset(ds, served_files, request.param == "remote")
    if request.param == "cloned":
        ds.repo.call_git(["update-server-info"])
        with local_server(ds.path) as origin_url:
            clone_ds = clone(origin_url, workpath / "clone")
            for dfile in served_files:
                clone_ds.repo.rm_url(dfile.path, dfile.url)
            yield (clone_ds, {df.path: df.content for df in served_files})
    else:
        yield (ds, {df.path: df.content for df in served_files})


@pytest.fixture(params=["remote", "local"])
def superdataset(served_files, request, tmp_home, tmp_path_factory):  # noqa: U100
    dspath = tmp_path_factory.mktemp("superdataset")
    ds = Dataset(dspath).create()
    sub = ds.create(dspath / "sub")
    initdataset(sub, served_files, request.param == "remote")
    return (ds, {os.path.join("sub", df.path): df.content for df in served_files})


@pytest.fixture
def big_url_dataset(tmp_home, tmp_path_factory):  # noqa: U100
    workpath = tmp_path_factory.mktemp("big_url_dataset")
    ds = Dataset(workpath / "ds").create()
    for path, url, _ in BIG_URLS:
        ds.repo.add_url_to_file(path, url, options=["--relaxed"])
    yield (ds, {path: digest for path, _, digest in BIG_URLS})



# -*- coding: utf-8 -*-
#
# datalad_metalad documentation build configuration file, created by
# sphinx-quickstart on Tue Oct 13 08:41:19 2015.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import datetime
import os
from os import pardir
from os.path import abspath, dirname, exists
from os.path import join as opj
import sys

import datalad_fuse

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

# generate missing pieces
for setup_py_path in (
    opj(pardir, "setup.py"),  # travis
    opj(pardir, pardir, "setup.py"),
):  # RTD
    if exists(setup_py_path):
        sys.path.insert(0, os.path.abspath(dirname(setup_py_path)))
        try:
            for cmd in ("manpage",):  #'examples':
                os.system(
                    "{} build_{} --cmdsuite {} --manpath {} --rstpath {}".format(
                        setup_py_path,
                        cmd,
                        "datalad_fuse:command_suite",
                        abspath(opj(dirname(setup_py_path), "build", "man")),
                        opj(dirname(__file__), "generated", "man"),
                    )
                )
        except Exception:
            # shut up and do your best
            pass

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.ifconfig",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
]

# for the module reference
autosummary_generate = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = u"Datalad Extension Template"
copyright = u"2018-{}, DataLad team".format(datetime.datetime.now().year)
author = u"DataLad team"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
version = datalad_fuse.__version__
release = version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {"https://docs.python.org/": None}

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/datalad_logo.png"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# If true, the index is split into individual pages for each letter.
html_split_index = True

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False



from __future__ import annotations

import pytest

from datalad_fuse.utils import AnnexDir, AnnexKey, is_annex_dir_or_key

SAMPLE_KEY = "MD5E-s1064--8804d3d11f17e33bd912f1f0947afdb9.json"
URL_KEY = "URL--http&c%%127.0.0.1&c55485%binary.png"

SAMPLE_ANNEX_KEY = AnnexKey(
    backend="MD5E",
    size=1064,
    name="8804d3d11f17e33bd912f1f0947afdb9",
    suffix=".json",
)

URL_ANNEX_KEY = AnnexKey(backend="URL", name="http://127.0.0.1:55485/binary.png")


@pytest.mark.parametrize(
    "path,expected",
    [
        (f".git/annex/objects/p0/4v/{SAMPLE_KEY}/{SAMPLE_KEY}", SAMPLE_ANNEX_KEY),
        (f".git/annex/objects/p2/pX/{URL_KEY}/{URL_KEY}", URL_ANNEX_KEY),
        (f".git/annex/objects/p0/4v/{SAMPLE_KEY}/", AnnexDir(".")),
        (f".git/annex/objects/p0/4v/{SAMPLE_KEY}", AnnexDir(".")),
        (".git/annex/objects/p0/4v", AnnexDir(".")),
        (
            f"some/project/.git/annex/objects/p0/4v/{SAMPLE_KEY}/{SAMPLE_KEY}",
            SAMPLE_ANNEX_KEY,
        ),
        ("some/project/.git/annex/objects/p0/4v", AnnexDir("some/project")),
        (
            f"/usr/src/project/.git/annex/objects/p0/4v/{SAMPLE_KEY}/{SAMPLE_KEY}",
            SAMPLE_ANNEX_KEY,
        ),
        ("/usr/src/project/.git/annex/objects/p0/4v", AnnexDir("/usr/src/project")),
        ("foo.txt", None),
        ("foo.git/annex/objects/p0/4v", None),
        ("some/project/.git/refs/heads", None),
        ("some/project/.git/annex", None),
        ("some/project/.git/annex/other", None),
        ("some/project/.git/annex/objects/layout_config.json", None),
        ("some/project/.git/annex/objects/p0/layout_config.json", None),
        ("some/project/.git/annex/objects/p0/4v/layout_config.json", None),
        (f"some/project/.git/annex/objects/p0/4v/{SAMPLE_KEY}/notmatchingkey", None),
        (
            f"some/project/.git/annex/objects/p0/4v/{SAMPLE_KEY}/{SAMPLE_KEY}/notmatchingkey",
            None,
        ),
        ("some/project/.git/annex/objects/p0/4v/notmatchingkey/notmatchingkey", None),
        (
            "some/project/.git/embedded/sub/.git/annex/objects/p0/4v/"
            f"{SAMPLE_KEY}/{SAMPLE_KEY}",
            SAMPLE_ANNEX_KEY,
        ),
        (
            "some/project/.git/embedded/sub/.git/annex/objects/p0/4v",
            AnnexDir("some/project/.git/embedded/sub"),
        ),
    ],
)
def test_is_annex_dir_or_key(path: str, expected: AnnexDir | AnnexKey | None) -> None:
    assert is_annex_dir_or_key(path) == expected



from typing import Any, Dict, Iterator, Optional

from datalad.distribution.dataset import (
    Dataset,
    EnsureDataset,
    datasetmethod,
    require_dataset,
)
from datalad.interface.base import Interface, build_doc
from datalad.interface.results import get_status_dict
from datalad.interface.utils import eval_results
from datalad.support.constraints import EnsureNone
from datalad.support.param import Parameter

from .fsspec import DatasetAdapter


@build_doc
class FsspecCacheClear(Interface):
    """
    Clear fsspec cache
    """

    _params_ = {
        "dataset": Parameter(
            args=("-d", "--dataset"),
            doc="""dataset to operate on.  If no dataset is given, an
            attempt is made to identify the dataset based on the current
            working directory.""",
            constraints=EnsureDataset() | EnsureNone(),
        ),
        "recursive": Parameter(
            args=("-r", "--recursive"),
            action="store_true",
            doc="Clear caches of subdatasets as well",
        ),
    }

    @staticmethod
    @datasetmethod(name="fsspec_cache_clear")
    @eval_results
    def __call__(
        dataset: Optional[Dataset] = None, recursive: bool = False
    ) -> Iterator[Dict[str, Any]]:
        ds = require_dataset(
            dataset, purpose="clear fsspec cache", check_installed=True
        )
        DatasetAdapter(ds.path).clear()
        yield get_status_dict(action="fsspec-cache-clear", ds=ds, status="ok")
        if recursive:
            for subds in ds.subdatasets(
                recursive=True, fulfilled=True, result_renderer="disabled"
            ):
                DatasetAdapter(subds["path"]).clear()
                yield get_status_dict(
                    action="fsspec-cache-clear",
                    refds=ds.path,
                    ds=Dataset(subds["path"]),
                    parentds=subds["parentds"],
                    status="ok",
                )



# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the DataLad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##


import datetime
import os

from os.path import (
    dirname,
    join as opj,
)
from setuptools import Command, DistutilsOptionError
from setuptools.config import read_configuration

import versioneer

from . import formatters as fmt


class BuildManPage(Command):
    # The BuildManPage code was originally distributed
    # under the same License of Python
    # Copyright (c) 2014 Oz Nahum Tiram  <nahumoz@gmail.com>

    description = 'Generate man page from an ArgumentParser instance.'

    user_options = [
        ('manpath=', None,
         'output path for manpages (relative paths are relative to the '
         'datalad package)'),
        ('rstpath=', None,
         'output path for RST files (relative paths are relative to the '
         'datalad package)'),
        ('parser=', None, 'module path to an ArgumentParser instance'
         '(e.g. mymod:func, where func is a method or function which return'
         'a dict with one or more arparse.ArgumentParser instances.'),
        ('cmdsuite=', None, 'module path to an extension command suite '
         '(e.g. mymod:command_suite) to limit the build to the contained '
         'commands.'),
    ]

    def initialize_options(self):
        self.manpath = opj('build', 'man')
        self.rstpath = opj('docs', 'source', 'generated', 'man')
        self.parser = 'datalad.cmdline.main:setup_parser'
        self.cmdsuite = None

    def finalize_options(self):
        if self.manpath is None:
            raise DistutilsOptionError('\'manpath\' option is required')
        if self.rstpath is None:
            raise DistutilsOptionError('\'rstpath\' option is required')
        if self.parser is None:
            raise DistutilsOptionError('\'parser\' option is required')
        mod_name, func_name = self.parser.split(':')
        fromlist = mod_name.split('.')
        try:
            mod = __import__(mod_name, fromlist=fromlist)
            self._parser = getattr(mod, func_name)(
                ['datalad'],
                formatter_class=fmt.ManPageFormatter,
                return_subparsers=True,
                # ignore extensions only for the main package to avoid pollution
                # with all extension commands that happen to be installed
                help_ignore_extensions=self.distribution.get_name() == 'datalad')

        except ImportError as err:
            raise err
        if self.cmdsuite:
            mod_name, suite_name = self.cmdsuite.split(':')
            mod = __import__(mod_name, fromlist=mod_name.split('.'))
            suite = getattr(mod, suite_name)
            self.cmdlist = [c[2] if len(c) > 2 else c[1].replace('_', '-')
                            for c in suite[1]]

        self.announce('Writing man page(s) to %s' % self.manpath)
        self._today = datetime.date.today()

    @classmethod
    def handle_module(cls, mod_name, **kwargs):
        """Module specific handling.

        This particular one does
        1. Memorize (at class level) the module name of interest here
        2. Check if 'datalad.extensions' are specified for the module,
           and then analyzes them to obtain command names it provides

        If cmdline commands are found, its entries are to be used instead of
        the ones in datalad's _parser.

        Parameters
        ----------
        **kwargs:
            all the kwargs which might be provided to setuptools.setup
        """
        cls.mod_name = mod_name

        exts = kwargs.get('entry_points', {}).get('datalad.extensions', [])
        for ext in exts:
            assert '=' in ext      # should be label=module:obj
            ext_label, mod_obj = ext.split('=', 1)
            assert ':' in mod_obj  # should be module:obj
            mod, obj = mod_obj.split(':', 1)
            assert mod_name == mod  # AFAIK should be identical

            mod = __import__(mod_name)
            if hasattr(mod, obj):
                command_suite = getattr(mod, obj)
                assert len(command_suite) == 2  # as far as I see it
                if not hasattr(cls, 'cmdline_names'):
                    cls.cmdline_names = []
                cls.cmdline_names += [
                    cmd
                    for _, _, cmd, _ in command_suite[1]
                ]

    def run(self):

        dist = self.distribution
        #homepage = dist.get_url()
        #appname = self._parser.prog
        appname = 'datalad'

        cfg = read_configuration(
            opj(dirname(dirname(__file__)), 'setup.cfg'))['metadata']

        sections = {
            'Authors': """{0} is developed by {1} <{2}>.""".format(
                appname, cfg['author'], cfg['author_email']),
        }

        for cls, opath, ext in ((fmt.ManPageFormatter, self.manpath, '1'),
                                (fmt.RSTManPageFormatter, self.rstpath, 'rst')):
            if not os.path.exists(opath):
                os.makedirs(opath)
            for cmdname in getattr(self, 'cmdline_names', list(self._parser)):
                if hasattr(self, 'cmdlist') and cmdname not in self.cmdlist:
                    continue
                p = self._parser[cmdname]
                cmdname = "{0}{1}".format(
                    'datalad ' if cmdname != 'datalad' else '',
                    cmdname)
                format = cls(
                    cmdname,
                    ext_sections=sections,
                    version=versioneer.get_version())
                formatted = format.format_man_page(p)
                with open(opj(opath, '{0}.{1}'.format(
                        cmdname.replace(' ', '-'),
                        ext)),
                        'w') as f:
                    f.write(formatted)


class BuildRSTExamplesFromScripts(Command):
    description = 'Generate RST variants of example shell scripts.'

    user_options = [
        ('expath=', None, 'path to look for example scripts'),
        ('rstpath=', None, 'output path for RST files'),
    ]

    def initialize_options(self):
        self.expath = opj('docs', 'examples')
        self.rstpath = opj('docs', 'source', 'generated', 'examples')

    def finalize_options(self):
        if self.expath is None:
            raise DistutilsOptionError('\'expath\' option is required')
        if self.rstpath is None:
            raise DistutilsOptionError('\'rstpath\' option is required')
        self.announce('Converting example scripts')

    def run(self):
        opath = self.rstpath
        if not os.path.exists(opath):
            os.makedirs(opath)

        from glob import glob
        for example in glob(opj(self.expath, '*.sh')):
            exname = os.path.basename(example)[:-3]
            with open(opj(opath, '{0}.rst'.format(exname)), 'w') as out:
                fmt.cmdline_example_to_rst(
                    open(example),
                    out=out,
                    ref='_example_{0}'.format(exname))


class BuildConfigInfo(Command):
    description = 'Generate RST documentation for all config items.'

    user_options = [
        ('rstpath=', None, 'output path for RST file'),
    ]

    def initialize_options(self):
        self.rstpath = opj('docs', 'source', 'generated', 'cfginfo')

    def finalize_options(self):
        if self.rstpath is None:
            raise DistutilsOptionError('\'rstpath\' option is required')
        self.announce('Generating configuration documentation')

    def run(self):
        opath = self.rstpath
        if not os.path.exists(opath):
            os.makedirs(opath)

        from datalad.interface.common_cfg import definitions as cfgdefs
        from datalad.dochelpers import _indent

        categories = {
            'global': {},
            'local': {},
            'dataset': {},
            'misc': {}
        }
        for term, v in cfgdefs.items():
            categories[v.get('destination', 'misc')][term] = v

        for cat in categories:
            with open(opj(opath, '{}.rst.in'.format(cat)), 'w') as rst:
                rst.write('.. glossary::\n')
                for term, v in sorted(categories[cat].items(), key=lambda x: x[0]):
                    rst.write(_indent(term, '\n  '))
                    qtype, docs = v.get('ui', (None, {}))
                    desc_tmpl = '\n'
                    if 'title' in docs:
                        desc_tmpl += '{title}:\n'
                    if 'text' in docs:
                        desc_tmpl += '{text}\n'
                    if 'default' in v:
                        default = v['default']
                        if hasattr(default, 'replace'):
                            # protect against leaking specific home dirs
                            v['default'] = default.replace(os.path.expanduser('~'), '~')
                        desc_tmpl += 'Default: {default}\n'
                    if 'type' in v:
                        type_ = v['type']
                        if hasattr(type_, 'long_description'):
                            type_ = type_.long_description()
                        else:
                            type_ = type_.__name__
                        desc_tmpl += '\n[{type}]\n'
                        v['type'] = type_
                    if desc_tmpl == '\n':
                        # we need something to avoid joining terms
                        desc_tmpl += 'undocumented\n'
                    v.update(docs)
                    rst.write(_indent(desc_tmpl.format(**v), '    '))



from itertools import islice
import os.path
import sys
from typing import Any, Dict, Iterator, Optional

from datalad.distribution.dataset import (
    Dataset,
    EnsureDataset,
    datasetmethod,
    require_dataset,
)
from datalad.interface.base import Interface, build_doc
from datalad.interface.results import get_status_dict
from datalad.interface.utils import eval_results
from datalad.support.constraints import EnsureInt, EnsureNone, EnsureStr
from datalad.support.param import Parameter

from .fsspec import FsspecAdapter

DEFAULT_LINES = 10


@build_doc
class FsspecHead(Interface):
    """
    Show leading lines/bytes of an annexed file by fetching its data from a
    remote URL
    """

    result_renderer = "tailored"

    _params_ = {
        "dataset": Parameter(
            args=("-d", "--dataset"),
            doc="""dataset to operate on.  If no dataset is given, an
            attempt is made to identify the dataset based on the current
            working directory.""",
            constraints=EnsureDataset() | EnsureNone(),
        ),
        "lines": Parameter(
            args=("-n", "--lines"),
            doc="How many lines to show (default 10)",
            constraints=EnsureInt() | EnsureNone(),
        ),
        "bytes": Parameter(
            args=("-c", "--bytes"),
            doc="How many bytes to show",
            constraints=EnsureInt() | EnsureNone(),
        ),
        "mode_transparent": Parameter(
            args=("--mode-transparent",),
            action="store_true",
            doc="Support reading from .git directory",
        ),
        "path": Parameter(
            args=("path",),
            doc="Path to an annexed file to show the leading contents of",
            constraints=EnsureStr(),
        ),
    }

    @staticmethod
    @datasetmethod(name="fsspec_head")
    @eval_results
    def __call__(
        path: str,
        dataset: Optional[Dataset] = None,
        lines: Optional[int] = None,
        bytes: Optional[int] = None,
        mode_transparent: bool = False,
    ) -> Iterator[Dict[str, Any]]:
        ds = require_dataset(dataset, purpose="fetch file data", check_installed=True)
        if lines is not None and bytes is not None:
            raise ValueError("'lines' and 'bytes' are mutually exclusive")
        elif lines is None and bytes is None:
            lines = DEFAULT_LINES
        with FsspecAdapter(ds.path, mode_transparent=mode_transparent) as fsa:
            if not os.path.isabs(path):
                path = os.path.join(ds.path, path)
            with fsa.open(path) as fp:
                if lines is not None:
                    blob = b"".join(islice(fp, lines))
                else:
                    blob = fp.read(bytes)
        yield get_status_dict(action="fsspec-head", ds=ds, status="ok", data=blob)

    @staticmethod
    def custom_result_renderer(res, **_):
        sys.stdout.buffer.write(res["data"])



#!/usr/bin/env python
import os.path
import sys

from setuptools import setup

# This is needed for versioneer to be importable when building with PEP 517.
# See <https://github.com/warner/python-versioneer/issues/193> and links
# therein for more information.
sys.path.append(os.path.dirname(__file__))

from _datalad_buildsupport.setup import BuildManPage  # noqa: E402
import versioneer  # noqa: E402

cmdclass = versioneer.get_cmdclass()
cmdclass.update(build_manpage=BuildManPage)

# Give setuptools a hint to complain if it's too old a version
# 30.3.0 allows us to put most metadata in setup.cfg
# Should match pyproject.toml
SETUP_REQUIRES = ["setuptools >= 30.3.0"]
# This enables setuptools to install wheel on-the-fly
SETUP_REQUIRES += ["wheel"] if "bdist_wheel" in sys.argv else []

if __name__ == "__main__":
    setup(
        name="datalad-fuse",
        version=versioneer.get_version(),
        cmdclass=cmdclass,
        setup_requires=SETUP_REQUIRES,
    )



"""DataLad FUSE extension"""

__docformat__ = "restructuredtext"

from typing import Any, Dict, Iterator, Optional

from datalad.distribution.dataset import (
    Dataset,
    EnsureDataset,
    datasetmethod,
    require_dataset,
)
from datalad.interface.base import Interface, build_doc
from datalad.interface.results import get_status_dict
from datalad.interface.utils import eval_results
from datalad.support.constraints import EnsureNone
from datalad.support.param import Parameter

from ._version import get_versions

# Defines a datalad command suite.
# This variable must be bound as a setuptools entrypoint
command_suite = (
    # description of the command suite, displayed in cmdline help
    "DataLad FUSE command suite",
    [
        # specification of a command, any number of commands can be defined
        (
            # importable module that contains the command implementation
            "datalad_fuse",
            # name of the command class implementation in above module
            "FuseFS",
            # optional name of the command in the cmdline API
            "fusefs",
            # optional name of the command in the Python API
            "fusefs",
        ),
        ("datalad_fuse.fsspec_head", "FsspecHead", "fsspec-head", "fsspec_head"),
        (
            "datalad_fuse.fsspec_cache_clear",
            "FsspecCacheClear",
            "fsspec-cache-clear",
            "fsspec_cache_clear",
        ),
    ],
)


# decoration auto-generates standard help
@build_doc
# all commands must be derived from Interface
class FuseFS(Interface):
    # first docstring line is used a short description in the cmdline help
    # the rest is put in the verbose help and manpage
    """
    FUSE File system providing transparent access to files under DataLad
    control
    """

    # parameters of the command, must be exhaustive
    _params_ = {
        "dataset": Parameter(
            args=("-d", "--dataset"),
            doc="""dataset to operate on.  If no dataset is given, an
                attempt is made to identify the dataset based on the current
                working directory.""",
            constraints=EnsureDataset() | EnsureNone(),
        ),
        "mount_path": Parameter(
            args=("mount_path",),
            metavar="PATH",
            doc="""Path where to mount the dataset (should exist).""",
        ),
        "allow_other": Parameter(
            args=("--allow-other",),
            action="store_true",
            doc=(
                "Allow all users to access files in the mount"
                " (Requires configuration in /etc/fuse.conf)"
            ),
        ),
        "foreground": Parameter(
            args=("-f", "--foreground"),
            action="store_true",
            doc="""Run process in foreground [required].""",
        ),
        "mode_transparent": Parameter(
            args=("--mode-transparent",),
            action="store_true",
            doc="Expose .git directory",
        ),
        # TODO: (might better become config vars?)
        # --cache=persist
        # --recursive=follow,get - encountering submodule might install it first
        # --git=[hide],show - hide .git in the FUSE space to avoid confusion/etc
    }

    @staticmethod
    @datasetmethod(name="fusefs")
    @eval_results
    def __call__(
        mount_path: str,
        dataset: Optional[Dataset] = None,
        foreground: bool = False,
        mode_transparent: bool = False,
        allow_other: bool = False,
    ) -> Iterator[Dict[str, Any]]:
        from fuse import FUSE

        from .fuse_ import DataLadFUSE

        if not foreground:
            yield get_status_dict(
                action="fusefs",
                path=mount_path,
                status="error",
                message="fusefs does not work properly without --foreground",
            )
            return
        ds = require_dataset(
            dataset, purpose="mount as FUSE system", check_installed=True
        )
        FUSE(
            DataLadFUSE(ds.path, mode_transparent=mode_transparent),
            mount_path,
            foreground=foreground,
            allow_other=allow_other,
        )
        yield get_status_dict(action="fusefs", path=mount_path, status="ok")


__version__ = get_versions()["version"]
del get_versions



from __future__ import annotations

from collections.abc import Iterator
from datetime import datetime, timezone
from enum import Enum
import logging
import os
import os.path
from pathlib import Path
from types import SimpleNamespace, TracebackType
from typing import IO, Any, Optional, Tuple, cast

import aiohttp
from aiohttp_retry import ListRetry, RetryClient
from datalad.distribution.dataset import Dataset
from datalad.support.annexrepo import AnnexRepo
from datalad.utils import get_dataset_root
from fsspec.exceptions import BlocksizeMismatchError
from fsspec.implementations.cached import CachingFileSystem
from fsspec.implementations.http import HTTPFileSystem
import methodtools

from .consts import CACHE_SIZE
from .utils import AnnexKey, is_annex_dir_or_key

lgr = logging.getLogger("datalad.fuse.fsspec")

FileState = Enum("FileState", "NOT_ANNEXED NO_CONTENT HAS_CONTENT")


class DatasetAdapter:
    def __init__(self, path: str | Path, mode_transparent: bool = False) -> None:
        self.path = Path(path)
        self.mode_transparent = mode_transparent
        ds = Dataset(path)
        self.annex: Optional[AnnexRepo]
        if isinstance(ds.repo, AnnexRepo):
            self.annex = ds.repo
        else:
            self.annex = None
        self.commit_dt = datetime.fromtimestamp(
            ds.repo.get_commit_date(), tz=timezone.utc
        )
        self.fs = CachingFileSystem(
            fs=HTTPFileSystem(get_client=get_client),
            # target_protocol='blockcache',
            cache_storage=os.path.join(path, ".git", "datalad", "cache", "fsspec"),
            # cache_check=600,
            # block_size=1024,
            # check_files=True,
            # expiry_times=True,
            # same_names=True
        )

    def close(self) -> None:
        if self.annex is not None:
            self.annex._batched.clear()

    @methodtools.lru_cache(maxsize=CACHE_SIZE)
    def get_file_state(self, relpath: str) -> tuple[FileState, Optional[AnnexKey]]:
        p = self.path / relpath
        lgr.debug("get_file_state: %s", relpath)

        def handle_path_under_annex_objects(
            p: Path,
        ) -> tuple[FileState, Optional[AnnexKey]]:
            iadok = is_annex_dir_or_key(p)
            if isinstance(iadok, AnnexKey):
                if p.exists():
                    return (FileState.HAS_CONTENT, iadok)
                else:
                    return (FileState.NO_CONTENT, iadok)
            else:
                return (FileState.NOT_ANNEXED, None)

        # Shortcut handling of content under .git, in particular - annex key paths
        if self.mode_transparent and relpath.startswith(".git/"):
            return handle_path_under_annex_objects(p)

        # A regular file or git link for which we need to explicitly ask annex about
        if not p.is_symlink():
            if p.stat().st_size < 1024 and self.annex is not None:
                if self.annex.is_under_annex(relpath, batch=True):
                    key = AnnexKey.parse(self.annex.get_file_key(relpath, batch=True))
                    if self.annex.file_has_content(relpath, batch=True):
                        return (FileState.HAS_CONTENT, key)
                    else:
                        return (FileState.NO_CONTENT, key)
            return (FileState.NOT_ANNEXED, None)

        return handle_path_under_annex_objects(
            Path(os.path.normpath(p.parent / os.readlink(p)))
        )

    def get_urls(self, key: str) -> Iterator[str]:
        assert self.annex is not None
        # TODO: switch to batch=True whenever
        # https://github.com/datalad/datalad/pull/6379 is merged/released.
        # Will need a recent git-annex to work!
        whereis = self.annex.whereis(key, output="full", batch=False, key=True)
        remote_uuids = []
        for ru, v in whereis.items():
            remote_uuids.append(ru)
            for u in v["urls"]:
                if is_http_url(u):
                    yield u

        path_mixed = self.annex._batched.get(
            "examinekey",
            annex_options=["--format=annex/objects/${hashdirmixed}${key}/${key}\\n"],
            path=self.annex.path,
        )(key)
        path_lower = self.annex._batched.get(
            "examinekey",
            annex_options=["--format=annex/objects/${hashdirlower}${key}/${key}\\n"],
            path=self.annex.path,
        )(key)

        uuid2remote_url = {}
        for r in self.annex.get_remotes():
            ru = self.annex.config.get(f"remote.{r}.annex-uuid")
            if ru is None:
                continue
            remote_url = self.annex.config.get(f"remote.{r}.url")
            if remote_url is None:
                continue
            remote_url = self.annex.config.rewrite_url(remote_url)
            uuid2remote_url[ru] = remote_url

        for ru in remote_uuids:
            try:
                base_url = uuid2remote_url[ru]
            except KeyError:
                continue
            if is_http_url(base_url):
                if base_url.lower().rstrip("/").endswith("/.git"):
                    paths = [path_mixed, path_lower]
                else:
                    paths = [
                        path_lower,
                        path_mixed,
                        f".git/{path_lower}",
                        f".git/{path_mixed}",
                    ]
                for p in paths:
                    yield base_url.rstrip("/") + "/" + p

    def open(
        self,
        relpath: str,
        mode: str = "rb",
        encoding: str = "utf-8",
        errors: Optional[str] = None,
    ) -> IO:
        if mode not in ("r", "rb", "rt"):
            raise NotImplementedError("Only modes 'r', 'rb', and 'rt' are supported")
        if mode == "rb":
            kwargs = {}
        else:
            kwargs = {"encoding": encoding, "errors": errors}
        fstate, key = self.get_file_state(relpath)
        if fstate is FileState.NOT_ANNEXED:
            lgr.debug("%s: not under annex", relpath)
        else:
            lgr.debug(
                "%s: under annex, %s content",
                relpath,
                "has" if fstate is FileState.HAS_CONTENT else "does not have",
            )
        if fstate is FileState.NO_CONTENT:
            lgr.debug("%s: opening via fsspec", relpath)
            for url in self.get_urls(str(key)):
                try:
                    lgr.debug("%s: Attempting to open via URL %s", relpath, url)
                    return self.fs.open(url, mode, **kwargs)  # type: ignore
                except BlocksizeMismatchError as e:
                    lgr.warning(
                        "%s: Blocksize mismatch: %s; deleting cached file and"
                        " re-opening",
                        relpath,
                        e,
                    )
                    self.fs.pop_from_cache(url)
                    return self.fs.open(url, mode, **kwargs)  # type: ignore
                except FileNotFoundError as e:
                    lgr.debug(
                        "Failed to open file %s at URL %s: %s", relpath, url, str(e)
                    )
            raise IOError(
                f"Could not find a usable URL for {relpath} within {self.path}"
            )
        else:
            lgr.debug("%s: opening directly", relpath)
            return open(self.path / relpath, mode, **kwargs)  # type: ignore

    def clear(self) -> None:
        self.fs.clear_cache()


class FsspecAdapter:
    def __init__(self, root: str | Path, mode_transparent: bool = False) -> None:
        self.root = Path(root)
        self.mode_transparent = mode_transparent
        self.datasets: dict[Path, DatasetAdapter] = {}

    def __enter__(self) -> FsspecAdapter:
        return self

    def __exit__(
        self,
        _exc_type: Optional[type[BaseException]],
        _exc_val: Optional[BaseException],
        _exc_tb: Optional[TracebackType],
    ) -> None:
        for ds in self.datasets.values():
            ds.close()
        self.datasets.clear()

    @methodtools.lru_cache(maxsize=CACHE_SIZE)
    # TODO: optimize "caching" more since for all files under the same directory
    # they all would belong to the same dataset
    def get_dataset_path(self, path: str | Path) -> Path:
        path = Path(self.root, path)
        dspath = get_dataset_root(path)
        if dspath is None:
            raise ValueError(f"Path not under DataLad: {path}")
        dspath = Path(dspath)
        assert isinstance(dspath, Path)
        try:
            dspath.relative_to(self.root)
        except ValueError:
            raise ValueError(f"Path not under root dataset: {path}")
        return dspath

    def resolve_dataset(self, filepath: str | Path) -> tuple[DatasetAdapter, str]:
        dspath = self.get_dataset_path(filepath)
        try:
            dsap = self.datasets[dspath]
        except KeyError:
            dsap = self.datasets[dspath] = DatasetAdapter(
                dspath, mode_transparent=self.mode_transparent
            )
        relpath = str(Path(filepath).relative_to(dspath))
        return dsap, relpath

    def open(
        self,
        filepath: str | Path,
        mode: str = "rb",
        encoding: str = "utf-8",
        errors: Optional[str] = None,
    ) -> IO:
        dsap, relpath = self.resolve_dataset(filepath)
        lgr.debug(
            "%s: path resolved to %s in dataset at %s", filepath, relpath, dsap.path
        )
        return dsap.open(relpath, mode=mode, encoding=encoding, errors=errors)

    def get_file_state(
        self, filepath: str | Path
    ) -> tuple[FileState, Optional[AnnexKey]]:
        dsap, relpath = self.resolve_dataset(filepath)
        return cast(Tuple[FileState, Optional[AnnexKey]], dsap.get_file_state(relpath))

    def is_under_annex(self, filepath: str | Path) -> bool:
        dsap, relpath = self.resolve_dataset(filepath)
        fstate, _ = dsap.get_file_state(relpath)
        return fstate is not FileState.NOT_ANNEXED

    def get_commit_datetime(self, filepath: str | Path) -> datetime:
        dsap, _ = self.resolve_dataset(filepath)
        return dsap.commit_dt


def is_http_url(s: str) -> bool:
    return s.lower().startswith(("http://", "https://"))


async def on_request_start(
    _session: aiohttp.ClientSession,
    trace_config_ctx: SimpleNamespace,
    params: aiohttp.TraceRequestStartParams,
) -> None:
    if trace_config_ctx.trace_request_ctx["current_attempt"] > 1:
        lgr.warning("Retrying request to %s", params.url)


async def get_client(**kwargs: Any) -> RetryClient:
    trace_config = aiohttp.TraceConfig()
    trace_config.on_request_start.append(on_request_start)
    return RetryClient(
        client_session=aiohttp.ClientSession(
            trace_configs=[trace_config],
            **kwargs,
        ),
        retry_options=ListRetry(timeouts=[1, 2, 6, 15, 36]),
    )



from __future__ import annotations

from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
import re
from typing import Optional

from datalad_fuse.consts import CACHE_SIZE


@dataclass
class AnnexKey:
    # <https://git-annex.branchable.com/internals/key_format/>
    backend: str
    name: str
    size: Optional[int] = None
    mtime: Optional[int] = None
    chunk_size: Optional[int] = None
    chunk_number: Optional[int] = None
    suffix: Optional[str] = None

    def __str__(self) -> str:
        s = self.backend
        if self.size is not None:
            s += f"-s{self.size}"
        if self.mtime is not None:
            s += f"-m{self.mtime}"
        if self.chunk_size is not None:
            s += f"-S{self.chunk_size}"
        if self.chunk_number is not None:
            s += f"-C{self.chunk_number}"
        s += f"--{self.name}"
        if self.suffix is not None:
            s += self.suffix
        return s

    @classmethod
    def parse(cls, s: str) -> AnnexKey:
        m = re.fullmatch(
            r"(?P<backend>[A-Z0-9_]{2,14})"
            r"(?:-s(?P<size>[0-9]+))?"
            r"(?:-m(?P<mtime>[0-9]+))?"
            r"(?:-S(?P<chunk_size>[0-9]+)-C(?P<chunk_number>[0-9]+))?"
            r"--(?P<name>.+)",
            s,
        )
        if m:
            backend = m["backend"]
            size = int(m["size"]) if m["size"] is not None else None
            mtime = int(m["mtime"]) if m["mtime"] is not None else None
            chunk_size = int(m["chunk_size"]) if m["chunk_size"] is not None else None
            chunk_number = (
                int(m["chunk_number"]) if m["chunk_number"] is not None else None
            )
            name = m["name"]
            if backend.endswith("E"):
                name, sep, suffix = name.rpartition(".")
                suffix = sep + suffix
            else:
                suffix = None
            return cls(
                backend=backend,
                size=size,
                mtime=mtime,
                chunk_size=chunk_size,
                chunk_number=chunk_number,
                name=name,
                suffix=suffix,
            )
        else:
            raise ValueError(f"invalid git-annex key: {s!r}")

    @classmethod
    def parse_filename(cls, s: str) -> AnnexKey:
        fields, sep, name = s.partition("--")
        # See `keyFile` and `fileKey` in `Annex/Locations.hs` in the git-annex
        # source
        name = (
            name.replace("%", "/")
            .replace("&c", ":")
            .replace("&s", "%")
            .replace("&a", "&")
        )
        return cls.parse(fields + sep + name)


@dataclass
class AnnexDir:
    topdir: str


# might be called twice in rapid succession for an annex key path
@lru_cache(maxsize=CACHE_SIZE)
def is_annex_dir_or_key(path: str | Path) -> AnnexDir | AnnexKey | None:
    parts = list(Path(path).parts)
    start = 0
    while True:
        try:
            i = parts.index(".git", start)
        except ValueError:
            return None
        if parts[i + 1 : i + 3] == ["annex", "objects"] and all(
            re.fullmatch(r"[A-Za-z0-9]{2}", p) for p in parts[i + 3 : i + 5]
        ):
            topdir = str(Path(*parts[:i]))
            depth = len(parts) - i
            if depth <= 5:  # have only two level of hash'ing directories
                return AnnexDir(topdir)
            try:
                key = AnnexKey.parse_filename(parts[i + 5])
            except ValueError:
                pass
            else:
                # note: key and its directory must match in name
                if depth == 7 and parts[-1] == parts[-2]:
                    return key
                elif depth == 6:
                    return AnnexDir(topdir)
        start = i + 1



from concurrent.futures import ThreadPoolExecutor, as_completed
from contextlib import contextmanager
import hashlib
import os.path
from pathlib import Path
import subprocess
from typing import Iterator, Union

from datalad.api import Dataset
import pytest

pytestmark = pytest.mark.libfuse


@contextmanager
def fusing(
    source_dir: Union[str, Path],
    mount_dir: Path,
    transparent: bool = False,
    wait: bool = False,
) -> Iterator[Path]:
    mount_dir.mkdir(parents=True, exist_ok=True)
    if transparent:
        opts = ["--mode-transparent"]
    else:
        opts = []
    p = subprocess.Popen(
        [
            "datalad",
            "fusefs",
            "-d",
            str(source_dir),
            "--foreground",
            *opts,
            str(mount_dir),
        ]
    )
    # Check that the command didn't fail immediately:
    with pytest.raises(subprocess.TimeoutExpired):
        p.wait(timeout=3)
    try:
        yield mount_dir
    finally:
        p.terminate()
        p.wait(timeout=3)
    if wait:
        p.wait()


@pytest.mark.parametrize("transparent", [False, True])
def test_fuse(tmp_path, transparent, url_dataset):
    ds, data_files = url_dataset
    if transparent:
        dots = [".datalad", ".git", ".gitattributes"]
    else:
        dots = [".datalad", ".gitattributes"]
    with fusing(ds.path, tmp_path, transparent=transparent) as mount:
        assert sorted(q.name for q in mount.iterdir()) == dots + sorted(data_files)
        for fname, blob in data_files.items():
            assert os.path.getsize(mount / fname) == len(blob)
            assert (mount / fname).read_bytes() == blob


@pytest.mark.parametrize("cache_clear", [None, "recursive", "visited"])
@pytest.mark.parametrize("transparent", [False, True])
def test_fuse_subdataset(tmp_path, superdataset, cache_clear, transparent, tmp_home):
    if cache_clear is not None:
        with (tmp_home / ".gitconfig").open("a") as fp:
            print("", file=fp)
            print('[datalad "fusefs"]', file=fp)
            print(f"cache-clear = {cache_clear}", file=fp)
    ds, data_files = superdataset
    if transparent:
        dots = [".datalad", ".git", ".gitattributes"]
    else:
        dots = [".datalad", ".gitattributes"]
    with fusing(ds.path, tmp_path, transparent=transparent, wait=True) as mount:
        assert sorted(q.name for q in mount.iterdir()) == dots + [".gitmodules", "sub"]
        assert sorted(q.name for q in (mount / "sub").iterdir()) == dots + sorted(
            os.path.relpath(fname, "sub") for fname in data_files
        )
        for fname, blob in data_files.items():
            assert (mount / fname).read_bytes() == blob
    cachedir = ds.pathobj / "sub" / ".git" / "datalad" / "cache" / "fsspec"
    is_local = (ds.pathobj / next(iter(data_files))).exists()
    if is_local and cache_clear != "recursive":
        assert not cachedir.exists()
    elif cache_clear is not None:
        assert list(cachedir.iterdir()) == []
    else:
        assert list(cachedir.iterdir()) != []


def test_fuse_transparent_hash_object(tmp_path):
    ds = Dataset(tmp_path / "ds").create()
    with fusing(ds.path, tmp_path / "mount", transparent=True) as mount:
        CONTENT = "This is test text.\n"
        r = subprocess.run(
            [
                "git",
                "-P",
                "--git-dir",
                str(mount / ".git"),
                "hash-object",
                "-w",
                "--stdin",
            ],
            cwd=mount,
            check=True,
            universal_newlines=True,
            input=CONTENT,
            stdout=subprocess.PIPE,
        )
        blobhash = r.stdout.strip()
        r = subprocess.run(
            [
                "git",
                "-P",
                "--git-dir",
                str(mount / ".git"),
                "hash-object",
                "-w",
                "--stdin",
            ],
            cwd=mount,
            check=True,
            universal_newlines=True,
            input=CONTENT,
            stdout=subprocess.PIPE,
        )
        assert r.stdout.strip() == blobhash
        r = subprocess.run(
            ["git", "-P", "--git-dir", str(mount / ".git"), "show", blobhash],
            cwd=mount,
            check=True,
            universal_newlines=True,
            stdout=subprocess.PIPE,
        )
        assert r.stdout == CONTENT
    r = subprocess.run(
        ["git", "-P", "show", blobhash],
        cwd=ds.path,
        check=True,
        universal_newlines=True,
        stdout=subprocess.PIPE,
    )
    assert r.stdout == CONTENT


def test_fuse_transparent_hash_object_subdataset(tmp_path):
    ds = Dataset(tmp_path / "ds").create()
    ds.create(tmp_path / "ds" / "sub")
    with fusing(ds.path, tmp_path / "mount", transparent=True) as mount:
        CONTENT = "This is test text.\n"
        r = subprocess.run(
            [
                "git",
                "-P",
                "--git-dir",
                str(mount / "sub" / ".git"),
                "hash-object",
                "-w",
                "--stdin",
            ],
            cwd=mount / "sub",
            check=True,
            universal_newlines=True,
            input=CONTENT,
            stdout=subprocess.PIPE,
        )
        blobhash = r.stdout.strip()
        r = subprocess.run(
            [
                "git",
                "-P",
                "--git-dir",
                str(mount / ".git"),
                "hash-object",
                "-w",
                "--stdin",
            ],
            cwd=mount,
            check=True,
            universal_newlines=True,
            input=CONTENT,
            stdout=subprocess.PIPE,
        )
        assert r.stdout.strip() == blobhash
        r = subprocess.run(
            ["git", "-P", "--git-dir", str(mount / "sub" / ".git"), "show", blobhash],
            cwd=mount / "sub",
            check=True,
            universal_newlines=True,
            stdout=subprocess.PIPE,
        )
        assert r.stdout == CONTENT
    r = subprocess.run(
        ["git", "-P", "show", blobhash],
        cwd=ds.pathobj / "sub",
        check=True,
        universal_newlines=True,
        stdout=subprocess.PIPE,
    )
    assert r.stdout == CONTENT


def test_fuse_lock(tmp_path):
    CONTENT = "This is test text.\n"
    ds = Dataset(tmp_path / "ds").create(cfg_proc="text2git")
    (tmp_path / "ds" / "text.txt").write_text(CONTENT)
    ds.save(message="Create text file")
    with fusing(ds.path, tmp_path / "mount", transparent=True) as mount:
        subprocess.run(
            ["git-annex", "smudge", "--clean", "--", "text.txt"],
            cwd=mount,
            check=True,
            input=CONTENT,
            universal_newlines=True,
        )


def test_fuse_git_status(tmp_path):
    CONTENT = "This is test text.\n"
    ds = Dataset(tmp_path / "ds").create(cfg_proc="text2git")
    (tmp_path / "ds" / "text.txt").write_text(CONTENT)
    ds.save(message="Create text file")
    with fusing(ds.path, tmp_path / "mount", transparent=True) as mount:
        r = subprocess.run(
            ["git", "status", "--porcelain"],
            cwd=mount,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        assert r.stdout == ""


def test_parallel_access(tmp_path, big_url_dataset):
    ds, data_files = big_url_dataset
    with fusing(ds.path, tmp_path) as mount:
        with ThreadPoolExecutor() as pool:
            futures = {
                pool.submit(sha256_file, mount / path): dgst
                for path, dgst in data_files.items()
            }
            for fut in as_completed(futures.keys()):
                assert fut.result() == futures[fut]


def sha256_file(path):
    dgst = hashlib.sha256()
    with open(path, "rb") as fp:
        for chunk in iter(lambda: fp.read(65535), b""):
            dgst.update(chunk)
    return dgst.hexdigest()
